#include "FuncBases.h"

// --- Prototypes -----------------------------------------

Real haarB1(const Vecd &x);

Real m2TriB1(const Vecd &x);
Real m2TriB2(const Vecd &x);
Real m2TriB3(const Vecd &x);
Real m3TriB4(const Vecd &x);
Real m3TriB5(const Vecd &x);
Real m3TriB6(const Vecd &x);

// standard basis functions, unnormalised

Real m2QuadB1(const Vecd &x);
Real m2QuadB2(const Vecd &x);
Real m2QuadB3(const Vecd &x);
Real m2QuadB4(const Vecd &x);
Real m3QuadB3(const Vecd &x);
Real m3QuadB6(const Vecd &x);
Real m3QuadB7(const Vecd &x);
Real m3QuadB8(const Vecd &x);
Real m3QuadB9(const Vecd &x);

// standard basis functions, normalised

Real m3QuadB3S(const Vecd &x);
Real m3QuadB7S(const Vecd &x);
Real m3QuadB8S(const Vecd &x);
Real m3QuadB6S(const Vecd &x);
Real m3QuadB9S(const Vecd &x);

// basis functions set up to be 1 at each corner of the quad.
// (I.e., can use basis coefficients directly for gouraud
// shading, without having to reconstruct the basis functions.

Real m2QuadB1C(const Vecd &x);
Real m2QuadB2C(const Vecd &x);
Real m2QuadB3C(const Vecd &x);
Real m2QuadB4C(const Vecd &x);
Real m3QuadB1C(const Vecd &x);
Real m3QuadB2C(const Vecd &x);
Real m3QuadB3C(const Vecd &x);
Real m3QuadB4C(const Vecd &x);
Real m3QuadB5C(const Vecd &x);
Real m3QuadB6C(const Vecd &x);
Real m3QuadB7C(const Vecd &x);
Real m3QuadB8C(const Vecd &x);
Real m3QuadB9C(const Vecd &x);


// --- Basis Functions ------------------------------------

// Haar... for any domain!

Real haarB1(const Vecd &x)
{
	return(1);
}

// Linear triangular basis

Real m2TriB1(const Vecd &x)
{
	return(1 - x[0] - x[1]);
}

Real m2TriB2(const Vecd &x)
{
	return(x[0]);
}

Real m2TriB3(const Vecd &x)
{
	return(x[1]);
}

// utils

#define MAKE_TRI_BF(NAME, EXPR) \
	Real NAME(const Vecd &x) \
	{ \
		Real a, b, c; \
		TriSetCorners(x, a, b, c); \
		return(EXPR); \
	}

Void TriSetCorners(const Vecd &x, Real &a, Real &b, Real &c)
{
	a = m2TriB1(x);
	b = m2TriB2(x);
	c = m2TriB3(x);
}
 
#define cube(x) (sqr(x) * (x))

// Quadratic triangular basis

MAKE_TRI_BF(m3TriB1, sqr(a))
MAKE_TRI_BF(m3TriB2, sqr(b))
MAKE_TRI_BF(m3TriB3, sqr(c))
MAKE_TRI_BF(m3TriB4, 2 * a * b)
MAKE_TRI_BF(m3TriB5, 2 * b * c)
MAKE_TRI_BF(m3TriB6, 2 * c * a)

// Quadratic triangular flatlet-equivalent basis

#if 0
MAKE_TRI_BF(m2FTriB1, a * (1 + b + c))
MAKE_TRI_BF(m2FTriB2, b * (1 + a + c))
MAKE_TRI_BF(m2FTriB3, c * (1 + a + b))
MAKE_TRI_BF(m2FTriB4, 2 * (a * b + b * c + c * a))
#else
MAKE_TRI_BF(m2FTriB1, a)
MAKE_TRI_BF(m2FTriB2, b)
MAKE_TRI_BF(m2FTriB3, c)
MAKE_TRI_BF(m2FTriB4, (a * b + b * c + c * a))
#endif

// Cubic triangular basis

MAKE_TRI_BF(m4TriB1, cube(a))
MAKE_TRI_BF(m4TriB2, cube(b))
MAKE_TRI_BF(m4TriB3, cube(c))
MAKE_TRI_BF(m4TriB4, 3 * sqr(a) * b)
MAKE_TRI_BF(m4TriB5, 3 * sqr(a) * c)
MAKE_TRI_BF(m4TriB6, 3 * sqr(b) * c)
MAKE_TRI_BF(m4TriB7, 3 * a * sqr(b))
MAKE_TRI_BF(m4TriB8, 3 * a * sqr(c))
MAKE_TRI_BF(m4TriB9, 3 * b * sqr(c))
MAKE_TRI_BF(m4TriB10, 6 * a * b * c)

// Cubic triangular flatlet-equivalent basis

#if 0
MAKE_TRI_BF(m3FTriB1, sqr(a) * (a + 3 * (b + c)))
MAKE_TRI_BF(m3FTriB2, sqr(b) * (b + 3 * (c + a)))
MAKE_TRI_BF(m3FTriB3, sqr(c) * (c + 3 * (a + b)))
MAKE_TRI_BF(m3FTriB4, 3 * a * b * (a + b + 2 * c))
MAKE_TRI_BF(m3FTriB5, 3 * b * c * (b + c + 2 * a))
MAKE_TRI_BF(m3FTriB6, 3 * c * a * (c + a + 2 * b))
MAKE_TRI_BF(m3FTriB7, 3 * a * (a * b + 2 * b * c + c * a))
MAKE_TRI_BF(m3FTriB8, 3 * b * (a * b + b * c + 2 * c * a))
MAKE_TRI_BF(m3FTriB9, 3 * c * (2 * a * b + b * c + c * a))
#else

MAKE_TRI_BF(m3FTriB1, sqr(a))
MAKE_TRI_BF(m3FTriB2, 3 * a * b)
MAKE_TRI_BF(m3FTriB3, sqr(b))

MAKE_TRI_BF(m3FTriB4, 3 * c * a)
MAKE_TRI_BF(m3FTriB5, 3 * b * c)
MAKE_TRI_BF(m3FTriB6, 3 * b * (a * b + b * c + 2 * c * a))

MAKE_TRI_BF(m3FTriB7, sqr(c))
MAKE_TRI_BF(m3FTriB8, 3 * c * (2 * a * b + b * c + c * a))
MAKE_TRI_BF(m3FTriB9, 3 * a * (a * b + 2 * b * c + c * a))

#endif


// Bilinear basis over a parallelogram

Real m2QuadB1(const Vecd &x)
{
	return(1.0);
}

Real m2QuadB2(const Vecd &x)
{
	return(x[0]);
}

Real m2QuadB3(const Vecd &x)
{
	return(x[1]);
}

Real m2QuadB4(const Vecd &x)
{
	return(x[0] * x[1]);
}

// Biquadratic basis over a parallelogram (extends m2QuadBX)

Real m3QuadB3(const Vecd &x)
{
	return(sqr(x[0]));
}

Real m3QuadB7(const Vecd &x)
{
	return(sqr(x[1]));
}

Real m3QuadB8(const Vecd &x)
{
	return(sqr(x[1]) * x[0]);
}

Real m3QuadB6(const Vecd &x)
{
	return(sqr(x[0]) * x[1]);
}

Real m3QuadB9(const Vecd &x)
{
	return(sqr(x[0]) * sqr(x[1]));
}

// Normalised version of m3QuadBX

Real m3QuadB3S(const Vecd &x)
{
	return(sqr(x[0]) - 1.0/3.0);
}

Real m3QuadB7S(const Vecd &x)
{
	return(sqr(x[1]) - 1.0/3.0);
}

Real m3QuadB8S(const Vecd &x)
{
	return((sqr(x[1]) - 1.0/3.0) * x[0]);
}

Real m3QuadB6S(const Vecd &x)
{
	return((sqr(x[0]) - 1.0/3.0) * x[1]);
}

Real m3QuadB9S(const Vecd &x)
{
	return((sqr(x[0]) - 1.0/3.0) * (sqr(x[1]) - 1.0/3.0));
}

// Bilinear basis for quad with weights at the vertices

Real m2QuadB1C(const Vecd &x)
{
	return((1 - x[0]) * (1 - x[1]) / 4.0);
}

Real m2QuadB2C(const Vecd &x)
{
	return((1 + x[0]) * (1 - x[1]) / 4.0);
}

Real m2QuadB3C(const Vecd &x)
{
	return((1 + x[0]) * (1 + x[1]) / 4.0);
}

Real m2QuadB4C(const Vecd &x)
{
	return((1 - x[0]) * (1 + x[1]) / 4.0);
}

// ditto for biquadratic

Real m3QuadB1C(const Vecd &x)
{
	return(x[0] * (1 - x[0]) * x[1] * (1 - x[1]) / 4.0);
}

Real m3QuadB2C(const Vecd &x)
{
	return((1 - x[0]) * (1 + x[0]) * x[1] * (1 - x[1]) / -2.0);
}

Real m3QuadB3C(const Vecd &x)
{
	return(x[0] * (1 + x[0]) * x[1] * (1 - x[1]) / -4.0);
}

Real m3QuadB4C(const Vecd &x)
{
	return(x[0] * (1 - x[0]) * (1 + x[1]) * (1 - x[1]) / -2.0);
}

Real m3QuadB5C(const Vecd &x)
{
	return((1 - x[0]) * (1 + x[0]) * (1 + x[1]) * (1 - x[1]));
}

Real m3QuadB6C(const Vecd &x)
{
	return(x[0] * (1 + x[0]) * (1 + x[1]) * (1 - x[1]) / 2.0);
}


Real m3QuadB7C(const Vecd &x)
{
	return(x[0] * (1 - x[0]) * x[1] * (1 + x[1]) / -4.0);
}

Real m3QuadB8C(const Vecd &x)
{
	return((1 - x[0]) * (1 + x[0]) * x[1] * (1 + x[1]) / 2.0);
}

Real m3QuadB9C(const Vecd &x)
{
	return(x[0] * (1 + x[0]) * x[1] * (1 + x[1]) / 4.0);
}

Real m4LineB4(const Vecd &x)
{
	return(sqr(x[0]) * x[0]);
}

// --- Basis definitions -------------------------------------

// 1D
static BasisFunction haarB[] =  	{haarB1};
static BasisFunction m2B[] =		{m2QuadB1, m2QuadB2};
static BasisFunction m3B[] =		{m2QuadB1, m2QuadB2, m3QuadB3S};
static BasisFunction m4B[] =		{m2QuadB1, m2QuadB2, m3QuadB3S, m4LineB4};

// 2D
static BasisFunction m2TriB[] =		{m2TriB1, m2TriB2, m2TriB3};
static BasisFunction m2QuadB[] =	{m2QuadB1, m2QuadB2, m2QuadB3, m2QuadB4};
static BasisFunction m2QuadBC[] =	{m2QuadB1C, m2QuadB2C, m2QuadB3C, m2QuadB4C};
static BasisFunction m3TriB[] =		{m3TriB1, m3TriB2, m3TriB3, m3TriB4, m3TriB5, m3TriB6};
static BasisFunction m3QuadB[] =	{m2QuadB1, m2QuadB2, m3QuadB3, m2QuadB3, m2QuadB4, m3QuadB6,
										m3QuadB7, m3QuadB8, m3QuadB9};
static BasisFunction m3QuadBC[] =	{m3QuadB1C, m3QuadB2C, m3QuadB3C,
										m3QuadB4C, m3QuadB5C, m3QuadB6C,
										m3QuadB7C, m3QuadB8C, m3QuadB9C};
static BasisFunction m3QuadBS[] =	{m2QuadB1, m2QuadB2, m3QuadB3S, m2QuadB3, m2QuadB4, m3QuadB6S,
									m3QuadB7S, m3QuadB8S, m3QuadB9S};
static BasisFunction m4TriB[] =		{m4TriB1, m4TriB2, m4TriB3, m4TriB4, m4TriB5, m4TriB6,
									 m4TriB7, m4TriB8, m4TriB9, m4TriB10};
									 
static BasisFunction m2FTriB[] =	{m2FTriB1, m2FTriB2,
									 m2FTriB3, m2FTriB4};
static BasisFunction m3FTriB[] =	{m3FTriB1, m3FTriB2, m3FTriB3, 
									 m3FTriB4, m3FTriB5, m3FTriB6,
									 m3FTriB7, m3FTriB8, m3FTriB9};

// format: function array, number of basis functions, dimensionality of basis functions

LinFB HAAR_LINE(haarB, 1, 1);
LinFB M2_LINE(m2B, 2, 1);
LinFB M3_LINE(m3B, 3, 1);
LinFB M4_LINE(m4B, 3, 1);

TriFB HAAR_TRI(haarB, 1, 2);
LinFB HAAR_QUAD(haarB, 1, 2);
TriFB M2_TRI(m2TriB, 3, 2);
LinFB M2_QUAD(m2QuadB, 4, 2);
LinFB M2_QUAD_V(m2QuadBC, 4, 2);
TriFB M3_TRI(m3TriB, 6, 2);
LinFB M3_QUAD(m3QuadBS, 9, 2);
LinFB M3_QUAD_V(m3QuadBC, 9, 2);
TriFB M4_TRI(m4TriB, 10, 2);

TriFB M2F_TRI(m2FTriB, 4, 2);
TriFB M3F_TRI(m3FTriB, 9, 2);


// --- Function bases ------------------------------


/*
	Layout of sub-functions:

	+---+---+				+
	|   |   |				|\
	| 2 | 3 |				| \
	|   |   |				|2 \
	+---+---+				+---\
	|   |   |				|\ 3|\
	| 0 | 1 |				| \ | \
	|   |   |				|0 \|1 \
	+---+---+				+---+---+
	
*/

Real FunctionBasis::Eval(Int i, const Vecd &x)
{
	return(f[i](x));
}

Vecd LinFB::ChildMap(const Vecd &x, Int i, Int order)
//	Samples basis function nb in the quadrant denoted by i, as shown above
//	for order = 2, function dimension = 2.
//	This code will work with orders higher than 2. 
{
	Int		j, k, n = FuncDim();
	Vecd	sx(n);
		
	// Map x to the sub quad denoted by i, j, and call b with these coordinates.
	
	for (k = 0; k < n; k++)
	{
		j = i % order;
		i = i / order;
		sx[k] = (x[k] + 2 * j - (order - 1)) / order;
	}
	
	return(sx);			
}

Vecd TriFB::ChildMap(const Vecd &x, Int i, Int order)
//	Samples b in the triangle quadrant denoted by i, as shown above
//	for order = 2.
//	This code will work with orders higher than 2, but currently only
//	for triangles, not tetrahedrons or above.
{
	Int		j;
	Vecd	sx(2);
	
	// Map x, y to the sub triangle denoted by i, j, and call b with these coordinates.
	// We use the flipping trick to cover all the sub triangles
	
	j = i % order;
	i = i / order;

	if (i + j < order)
	// A normal sub triangle
	{
		sx[0] = (x[0] + j) / order;
		sx[1] = (x[1] + i) / order;

		return(sx);
	}
	else
	// A flipped sub triangle
	{
		sx[0] = (order - i - x[0]) / order;
		sx[1] = (order - j - x[1]) / order;
		// Note the flipping of x[0] and x[1] above. This is because if
		// we reflect both, the central triangle will have a backwards
		// normal in the coord system of our radiosity program.

		return(sx);
	}
}


