#ifndef __FunctionBasis__
#define __FunctionBasis__

#include "VLd.h"

typedef Real (*BasisFunction)(const Vecd &x);
// returns value of basis function sampled at point x

enum DomainType
{
	domLinear,
	domTriangular
};

class FunctionBasis
// represents a set of basis functions: [b0, b1, b2... ]
{
public:
	FunctionBasis(BasisFunction *fArray, Int n, Int d) : 
		f(fArray), n(n), d(d) {};

	virtual Real	Eval(Int i, const Vecd &x);
	// returns bi(x)
	
	virtual Int		Dim() { return n; };
	// How many basis functions in this basis.
	virtual Int 	FuncDim() { return d; };
	// 1 == line, 2 == area, etc.

	virtual Vecd 	ChildMap(const Vecd &x, Int i, Int order) = 0;
	
protected:
	BasisFunction	*f;
	Int				n;		//	How many basis functions.
	Int				d;		//	dimensionality of the basis functions.
};

class LinFB : public FunctionBasis
{
public:
	LinFB(BasisFunction *fArray, Int n, Int d) : FunctionBasis(fArray, n, d) {};
	Vecd 			ChildMap(const Vecd &x, Int i, Int order);
};

class TriFB : public FunctionBasis
{
public:
	TriFB(BasisFunction *fArray, Int n, Int d) : FunctionBasis(fArray, n, d) {};
	Vecd 			ChildMap(const Vecd &x, Int i, Int order);
};

class TensorFB : public FunctionBasis
{
//	Constructs a tensor basis with dimension d ^ m
//	from the source basis.
public:
	TensorFB(FunctionBasis *sourceBasis, Int m);

	Real Eval(Int i, const Vecd &x);
};


extern LinFB HAAR_LINE;	// Haar basis on a line
extern LinFB M2_LINE;	// M2 basis on a line
extern LinFB M3_LINE;	// M3 basis on a line
extern LinFB M4_LINE;	// M4 basis on a line

extern TriFB HAAR_TRI;	// Haar (constant) triangular basis (1 member)
extern TriFB M2_TRI;	// linear triangular basis (3 members)
extern TriFB M3_TRI;	// quadratic triangular basis (6 members)
extern TriFB M4_TRI;	// cubic triangular basis (10 members)
extern TriFB M2F_TRI;	// linear triangular basis + center function (4 members)
						// smooth equivalent of F2_Tri
extern TriFB M3F_TRI;	// quadratic triangular basis + 3 center functions (9 members)
						// smooth equivalent of F3_Tri

extern LinFB HAAR_QUAD;	// Haar (constant) quad basis (1 member)
extern LinFB M2_QUAD;	// bi-linear quad basis: st, s(1-t), etc. (3x3). (2x2)
extern LinFB M2_QUAD_V;	// ditto, but basis functions are 1 at each vertex
extern LinFB M3_QUAD;	// bi-quadratic quad basis: s^2t^2, s^2t(1-t), etc. (3x3)
extern LinFB M3_QUAD_N;	// bi-quadratic quad basis, normalised (3x3)
extern LinFB M3_QUAD_V;	// bi-quadratic quad basis, basis functions are 1 at each vertex (3x3)

#endif
